/* Screen Shield */

.unlock-dialog-clock {
  color: white;
  font-weight: 300;
  text-align: center;
  spacing: 24px;
  padding-bottom: 2.5em;
}

.unlock-dialog-clock-time {
  font-size: 64pt;
  padding-top: 42px;
  font-feature-settings: "tnum";
}

.unlock-dialog-clock-date {
  font-size: 16pt;
  font-weight: normal;
}

.unlock-dialog-clock-hint {
  font-weight: normal;
  padding-top: 48px;
}

.unlock-dialog-notifications-container {
  margin: 12px 0;
  spacing: 6px;
  width: 23em;
  background-color: transparent;
  .summary-notification-stack-scrollview {
    padding-top: 0;
    padding-bottom: 0;
  }

  .notification,
  .unlock-dialog-notification-source {
    padding: 12px 6px;
    border: none;
    background-color: transparentize($osd_bg_color,0.7);
    color: $osd_fg_color;
    border-radius: $modal_radius;

    &.critical { background-color: transparentize($osd_bg_color,0.1) }
  }
}

.unlock-dialog-notification-label {
  padding: 0px 0px 0px 12px;
}

.unlock-dialog-notification-count-text {
  weight: bold;
  padding: 0 6px;
  color: $osd_bg_color;
  background-color: transparentize($osd_fg_color, 0.7);
  border-radius: 99px;
  margin-right: 12px;

}

.screen-shield-background { //just the shadow, really
  background: black;
  box-shadow: 0px 2px 4px rgba(0,0,0,0.6);
}

#lockDialogGroup {
  background-color: lighten(#2e3436, 8%);
}

#unlockDialogNotifications {
  StButton#vhandle, StButton#hhandle {
    background-color: transparentize($bg_color,0.7);
    &:hover, &:focus { background-color: transparentize($bg_color,0.5); }
    &:active { background-color: transparentize($selected_bg_color,0.5); }
  }
}
