/* Dash */

$dash_placeholder_size: 32px;
$dash_spacing: $base_padding + 4px;
$dash_border_radius: $modal_radius * 1.5;

#dash {
  @extend %overview_panel;
  @include fontsize($base_font_size - 2);
  padding: ($dash_spacing / 2) 0;

  border-radius: 0 $dash_border_radius $dash_border_radius 0; 
  border-left-width: 0;
  &:rtl {
    border-radius: $dash_border_radius 0 0 $dash_border_radius;
    border-right-width: 0;
  }

  .placeholder {
    // background-image: url("resource:///org/gnome/shell/theme/dash-placeholder.svg");
    background-image:none;
    background-size: contain;
    height: $dash_placeholder_size;
  }

  .empty-dash-drop-target {
    width: $dash_placeholder_size;
    height: $dash_placeholder_size;
  }
}

// Dash Items
.dash-item-container > StWidget {
  padding: ($dash_spacing / 2) $dash_spacing;
}

// OSD Tooltip
.dash-label {
  background-color: transparentize($osd_bg_color,0.05);
  border-radius: $base_border_radius + 2px;
  border:none;
  box-shadow:0 0 0 1px $osd_outer_borders_color;
  color: $osd_fg_color;
  padding: $base_padding $base_padding + 2px;
  text-align: center;
  -x-offset: $base_margin * 2; // distance from the dash edge
}

// Show apps button
.show-apps {
  @include overview-icon($osd_fg_color);

  &:focus,
  &:checked {
    .overview-icon {
      background-color: darken($osd_bg_color,10%);
      color: $fg_color;
    }
  }
}

