// Links
.shell-link {
  color: $link_color;

  &:hover {
    color: lighten($link_color, 10%);
  }
}

// Outline for low res icons
.lowres-icon {
  icon-shadow: 0 1px 2px rgba(black, 0.3);
}

// Dropshadow for large icons
.icon-dropshadow {
  icon-shadow: 0 1px 2px rgba(black, 0.4);
}
